module Eval where

import Parser
import Research

import Data.List (partition, (\\))
import Data.Map  (Map)
import qualified Data.Map as Map
import qualified Data.Set as Set

evalSingle :: Env -> TricuAST -> Env
evalSingle env term
  | SDef name [] body <- term
  = case Map.lookup name env of
      Just existingValue
        | existingValue == evalAST env body -> env
        | otherwise -> errorWithoutStackTrace $
            "Unable to rebind immutable identifier: " ++ name
      Nothing ->
        let res = evalAST env body
        in Map.insert "!result" res (Map.insert name res env)
  | SApp func arg <- term
  = let res = apply (evalAST env func) (evalAST env arg)
      in Map.insert "!result" res env
  | SVar name <- term
  = case Map.lookup name env of
        Just v  -> Map.insert "!result" v env
        Nothing ->
          errorWithoutStackTrace $ "Variable `" ++ name ++ "` not defined\n\
          \This error should never occur here. Please report this as an issue."
  | otherwise
  = Map.insert "!result" (evalAST env term) env

evalTricu :: Env -> [TricuAST] -> Env
evalTricu env x = go env (reorderDefs env x)
  where
    go env []     = env
    go env [x]    =
      let updatedEnv = evalSingle env x
      in Map.insert "!result" (result updatedEnv) updatedEnv
    go env (x:xs) =
      evalTricu (evalSingle env x) xs

evalAST :: Env -> TricuAST -> T
evalAST env term
  | SLambda _ _ <- term = evalAST env (elimLambda term)
  | SVar   name <- term = evalVar name
  | TLeaf       <- term = Leaf
  | TStem  t    <- term = Stem (evalAST env t)
  | TFork  t u  <- term = Fork (evalAST env t) (evalAST env u)
  | SApp   t u  <- term = apply (evalAST env t) (evalAST env u)
  | SStr   s    <- term = ofString s
  | SInt   n    <- term = ofNumber n
  | SList  xs   <- term = ofList (map (evalAST env) xs)
  | SEmpty      <- term = Leaf
  | otherwise           = errorWithoutStackTrace "Unexpected AST term"
    where
      evalVar name = Map.findWithDefault
        (errorWithoutStackTrace $ "Variable " ++ name ++ " not defined")
        name env

elimLambda :: TricuAST -> TricuAST
elimLambda = go
  where
    go term
      | etaReduction term   = elimLambda $ etaReduceResult term
      | triagePattern term  = _TRI
      | composePattern term = _B
      | lambdaList term     = elimLambda $ lambdaListResult term
      | nestedLambda term   = nestedLambdaResult term
      | application term    = applicationResult term
      | otherwise           = term

    etaReduction (SLambda [v] (SApp f (SVar x))) = v == x && not (isFree v f)
    etaReduction _ = False
    etaReduceResult (SLambda [_] (SApp f _)) = f

    triagePattern (SLambda [a] (SLambda [b] (SLambda [c] body))) = body == triageBody a b c
    triagePattern _ = False

    composePattern (SLambda [f] (SLambda [g] (SLambda [x] body))) = body == composeBody f g x
    composePattern _ = False

    lambdaList (SLambda [_] (SList _)) = True
    lambdaList _ = False
    lambdaListResult (SLambda [v] (SList xs)) = SLambda [v] (foldr wrapTLeaf TLeaf xs)
    wrapTLeaf m r = SApp (SApp TLeaf m) r

    nestedLambda (SLambda (_:_) _) = True
    nestedLambda _ = False
    nestedLambdaResult (SLambda (v:vs) body)
      | null vs   = toSKI v (elimLambda body)
      | otherwise = elimLambda (SLambda [v] (SLambda vs body))

    application (SApp _ _) = True
    application _ = False
    applicationResult (SApp f g) = SApp (elimLambda f) (elimLambda g)

    toSKI x (SVar y)
      | x == y           = _I
      | otherwise        = SApp _K (SVar y)
    toSKI x t@(SApp n u)
      | not (isFree x t) = SApp _K t
      | otherwise        = SApp (SApp _S (toSKI x n)) (toSKI x u)
    toSKI x (SList xs)
      | not (isFree x (SList xs)) = SApp _K (SList xs)
      | otherwise = SList (map (toSKI x) xs)
    toSKI x t
      | not (isFree x t) = SApp _K t
      | otherwise        = errorWithoutStackTrace "Unhandled toSKI conversion"

    -- Combinators and special forms
    _S   = parseSingle "t (t (t t t)) t"
    _K   = parseSingle "t t"
    _I   = parseSingle "t (t (t t)) t"
    _B   = parseSingle "t (t (t t (t (t (t t t)) t))) (t t)"
    _TRI = parseSingle "t (t (t t (t (t (t t t))))) t"
    
    -- Pattern bodies
    triageBody a b c = SApp (SApp TLeaf (SApp (SApp TLeaf (SVar a)) (SVar b))) (SVar c)
    composeBody f g x = SApp (SVar f) (SApp (SVar g) (SVar x))

isFree :: String -> TricuAST -> Bool
isFree x = Set.member x . freeVars

freeVars :: TricuAST -> Set.Set String
freeVars (SVar    v    ) = Set.singleton v
freeVars (SList   s    ) = foldMap freeVars s
freeVars (SLambda v b  ) = foldr Set.delete (freeVars b) v
freeVars (SApp    f a  ) = freeVars f <> freeVars a
freeVars (TFork   l r  ) = freeVars l <> freeVars r
freeVars (SDef   _ _ b)  = freeVars b
freeVars (TStem   t    ) = freeVars t
freeVars (SInt    _    ) = Set.empty
freeVars (SStr    _    ) = Set.empty
freeVars  TLeaf          = Set.empty
freeVars  _              = Set.empty

reorderDefs :: Env -> [TricuAST] -> [TricuAST]
reorderDefs env defs
  | not (null missingDeps) =
      errorWithoutStackTrace $
        "Missing dependencies detected: " ++ show missingDeps
  | otherwise = orderedDefs ++ others
  where
    (defsOnly, others) = partition isDef defs
    defNames = [ name | SDef name _ _ <- defsOnly ]

    defsWithFreeVars = [(def, freeVars body) | def@(SDef _ _ body) <- defsOnly]

    graph = buildDepGraph defsOnly
    sortedDefs = sortDeps graph
    defMap = Map.fromList [(name, def) | def@(SDef name _ _) <- defsOnly]
    orderedDefs = map (defMap Map.!) sortedDefs

    freeVarsDefs = foldMap snd defsWithFreeVars
    freeVarsOthers = foldMap freeVars others
    allFreeVars = freeVarsDefs <> freeVarsOthers
    validNames = Set.fromList defNames `Set.union` Set.fromList (Map.keys env)
    missingDeps = Set.toList (allFreeVars `Set.difference` validNames)

    isDef SDef {} = True
    isDef _       = False

buildDepGraph :: [TricuAST] -> Map.Map String (Set.Set String)
buildDepGraph topDefs
  | not (null conflictingDefs) =
      errorWithoutStackTrace $
        "Conflicting definitions detected: " ++ show conflictingDefs
  | otherwise =
      Map.fromList
        [ (name, depends topDefs (SDef name [] body))
        | SDef name _ body <- topDefs]
  where
    defsMap = Map.fromListWith (++)
      [(name, [(name, body)]) | SDef name _ body <- topDefs]

    conflictingDefs =
      [ name
      | (name, defs) <- Map.toList defsMap
      , let bodies = map snd defs
      , not $ all (== head bodies) (tail bodies)
      ]

sortDeps :: Map.Map String (Set.Set String) -> [String]
sortDeps graph = go [] Set.empty (Map.keys graph)
  where
    go sorted sortedSet [] = sorted
    go sorted sortedSet remaining =
      let ready = [ name | name <- remaining
                        , let deps = Map.findWithDefault Set.empty name graph
                        , Set.isSubsetOf deps sortedSet ]
          notReady = remaining \\ ready
      in if null ready
         then errorWithoutStackTrace
          "ERROR: Cyclic dependency detected and prohibited.\n\
          \RESOLVE: Use nested lambdas."
         else go (sorted ++ ready)
                 (Set.union sortedSet (Set.fromList ready))
                 notReady

depends :: [TricuAST] -> TricuAST -> Set.Set String
depends topDefs (SDef _ _ body) =
  Set.intersection
    (Set.fromList [n | SDef n _ _ <- topDefs])
    (freeVars body)
depends _ _ = Set.empty

result :: Env -> T
result r = case Map.lookup "!result" r of
  Just a -> a
  Nothing -> errorWithoutStackTrace "No !result field found in provided env"

mainResult :: Env -> T
mainResult r = case Map.lookup "main" r of
  Just  a -> a
  Nothing -> errorWithoutStackTrace "No valid definition for `main` found."
